package com.hz.pm.api.meta.helper.basic;

import com.github.benmanes.caffeine.cache.Caffeine;
import com.github.benmanes.caffeine.cache.LoadingCache;
import com.hz.pm.api.meta.assembler.MetaTagAssembler;
import com.hz.pm.api.meta.helper.TagCache;
import com.hz.pm.api.meta.model.dto.TagDTO;
import com.hz.pm.api.meta.model.dto.TagTreeDTO;
import com.hz.pm.api.meta.service.IMetaTagService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.factory.InitializingBean;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author liuxinxin
 * @date 2022/12/22 上午11:56
 */
@Slf4j
@RequiredArgsConstructor
public abstract class AbstractTagsCache implements InitializingBean, TagCache {

    protected static final int MAX_LEVEL = 4;

    private final IMetaTagService metaTagService;

    protected LoadingCache<String, List<TagTreeDTO>> tagsCache;

    protected Map<String, TagDTO> tagDtoMap;

    @Override
    public void afterPropertiesSet() {
        tagsCache = Caffeine.newBuilder().expireAfterWrite(10, TimeUnit.MINUTES)
                .refreshAfterWrite(5, TimeUnit.MINUTES)
                .build(key -> {
                    // 查询全部
                    List<TagDTO> tagDTOList = metaTagService.queryAll();
                    if (tagDTOList.isEmpty()) {
                        log.warn("当前层级对应树不存在：{}", key);
                        return Collections.emptyList();
                    }
                    tagDtoMap = buildCateMap(tagDTOList);
                    List<TagDTO> tagInLevel = tagDTOList.stream().filter(tagDto -> {
                        // 只过滤出小于等于level的tag
                        if (Objects.isNull(tagDto.getTagLevel())) {
                            return false;
                        }
                        return Integer.parseInt(key) >= tagDto.getTagLevel();
                    }).collect(Collectors.toList());
                    List<TagTreeDTO> treeDtos = MetaTagAssembler.toTagTreeDTOList(tagInLevel);
                    return toTreeStructure(treeDtos);
                });
        warmUp();
    }

    private Map<String, TagDTO> buildCateMap(List<TagDTO> tagDTOList) {
        Map<String, TagDTO> tagDtoMapTemp = new ConcurrentHashMap<>(1024);
        tagDTOList.forEach(tagDTO -> tagDtoMapTemp.put(tagDTO.getTagCode(), tagDTO));
        return tagDtoMapTemp;
    }

    private List<TagTreeDTO> toTreeStructure(List<TagTreeDTO> rootList) {
        List<TagTreeDTO> nodeList = new ArrayList<>();
        for (TagTreeDTO tagTreeDTO : rootList) {
            if ("-1".equals(tagTreeDTO.getParentCode()) || Objects.isNull(tagTreeDTO.getParentCode())) {
                nodeList.add(tagTreeDTO);
            }
            tagTreeDTO.setChildren(getChildren(tagTreeDTO.getTagCode(), rootList));
        }
        return nodeList;
    }

    public void warmUp() {
        for (int level = 1; level <= MAX_LEVEL; level++) {
            tagsCache.get(level + "");
        }
    }

    protected List<TagTreeDTO> getChildren(String tagCode, List<TagTreeDTO> list) {
        List<TagTreeDTO> childList = new ArrayList<>();
        for (TagTreeDTO tagTreeDTO : list) {
            if (tagCode.equals(tagTreeDTO.getParentCode())) {
                childList.add(tagTreeDTO);
            }
        }
        for (TagTreeDTO tagTreeDTO : childList) {
            tagTreeDTO.setChildren(getChildren(tagTreeDTO.getTagCode(), list));
        }
        if (CollectionUtils.isEmpty(childList)) {
            return Collections.emptyList();
        }
        return childList;
    }

}
